package ru.dbotthepony.mc.otm.container

import it.unimi.dsi.fastutil.ints.IntCollection
import it.unimi.dsi.fastutil.ints.IntOpenHashSet
import net.minecraft.core.HolderLookup.Provider
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.NbtOps
import net.minecraft.nbt.Tag
import net.minecraft.resources.RegistryOps
import net.minecraft.world.SimpleContainer
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.ItemStack
import net.neoforged.neoforge.common.util.INBTSerializable
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.collect.iterateClearBits
import ru.dbotthepony.kommons.collect.iterateSetBits
import ru.dbotthepony.mc.otm.container.slotted.SlottedContainer
import ru.dbotthepony.mc.otm.core.collect.IntRange2Set
import ru.dbotthepony.mc.otm.core.collect.map
import ru.dbotthepony.mc.otm.core.isNotEmpty
import ru.dbotthepony.mc.otm.core.nbt.set
import java.util.BitSet

/**
 * Flexible base implementation of [IEnhancedContainer], designed to be inherited, or used as-is
 * if no specific logic is required (this implementation is more efficient than one provided by [SlottedContainer.simple]).
 *
 * This is supposed to be counterpart to [SimpleContainer] of Minecraft itself, with more features
 * and improved performance (inside [IEnhancedContainer] defined methods).
 */
abstract class EnhancedContainer<out S : IContainerSlot>(private val size: Int) : IEnhancedContainer<S>, INBTSerializable<CompoundTag> {
	private val items = Array(size) { ItemStack.EMPTY }
	private val observedItems = Array(size) { ItemStack.EMPTY }
	private val bitmap = BitSet(size)

	final override fun isEmpty(): Boolean {
		return bitmap.isEmpty
	}

	final override fun nextEmptySlot(startIndex: Int): Int {
		if (startIndex >= size)
			return -1
		else if (startIndex < 0)
			return bitmap.nextClearBit(0)
		else
			return bitmap.nextClearBit(startIndex)
	}

	final override fun nextNonEmptySlot(startIndex: Int): Int {
		if (startIndex >= size)
			return -1
		else if (startIndex < 0)
			return bitmap.nextSetBit(0)
		else
			return bitmap.nextSetBit(startIndex)
	}

	final override fun iterator(): Iterator<ItemStack> {
		return bitmap.iterateSetBits(size).map { this[it] }
	}

	final override fun nonEmptySlotIndexIterator(): IntIterator {
		return bitmap.iterateSetBits(size)
	}

	final override fun emptySlotIndexIterator(): IntIterator {
		return bitmap.iterateClearBits(size)
	}

	final override fun emptySlotIndexIterator(allowedSlots: IntCollection): IntIterator {
		if (allowedSlots is IntRange2Set && allowedSlots.isNotEmpty())
			return bitmap.iterateClearBits(allowedSlots.firstInt(), allowedSlots.lastInt() + 1)

		return super.emptySlotIndexIterator(allowedSlots)
	}

	final override fun nonEmptySlotIndexIterator(allowedSlots: IntCollection): IntIterator {
		if (allowedSlots is IntRange2Set && allowedSlots.isNotEmpty())
			return bitmap.iterateSetBits(allowedSlots.firstInt(), allowedSlots.lastInt() + 1)

		return super.nonEmptySlotIndexIterator(allowedSlots)
	}

	protected open fun notifySlotChanged(slot: Int, old: ItemStack) {}

	private fun observeSlotChanges(slot: Int) {
		if (items[slot].count != observedItems[slot].count || !ItemStack.isSameItemSameComponents(items[slot], observedItems[slot])) {
			notifySlotChanged(slot, observedItems[slot])

			if (items[slot].isEmpty) {
				items[slot] = ItemStack.EMPTY
				observedItems[slot] = ItemStack.EMPTY
				bitmap[slot] = false
			} else {
				notifySlotChanged(slot, observedItems[slot])
				observedItems[slot] = items[slot].copy()
				bitmap[slot] = true
			}
		}
	}

	override fun setChanged(slot: Int) {
		observeSlotChanges(slot)
	}

	override fun clearContent() {
		items.fill(ItemStack.EMPTY)
		observedItems.fill(ItemStack.EMPTY)
		bitmap.clear()
	}

	override fun setChanged() {
		for (slot in 0 until size)
			observeSlotChanges(slot)
	}

	final override fun getContainerSize(): Int {
		return size
	}

	final override fun getItem(slot: Int): ItemStack {
		return items[slot]
	}

	final override fun removeItem(slot: Int, amount: Int): ItemStack {
		val item = items[slot]

		if (item.isEmpty)
			return ItemStack.EMPTY
		else if (item.count >= amount)
			return removeItemNoUpdate(slot)
		else {
			val split = item.split(amount)
			notifySlotChanged(slot, observedItems[slot])
			observedItems[slot] = item.copy()
			return split
		}
	}

	final override fun removeItemNoUpdate(slot: Int): ItemStack {
		val item = items[slot]

		if (item.isEmpty)
			return ItemStack.EMPTY

		items[slot] = ItemStack.EMPTY
		notifySlotChanged(slot, observedItems[slot])
		observedItems[slot] = ItemStack.EMPTY

		return item
	}

	final override fun setItem(slot: Int, itemStack: ItemStack) {
		if (itemStack.count != items[slot].count || !ItemStack.isSameItemSameComponents(itemStack, items[slot])) {
			items[slot] = itemStack
			notifySlotChanged(slot, observedItems[slot])
			observedItems[slot] = itemStack.copy()
		}
	}

	override fun stillValid(player: Player): Boolean {
		return true
	}

	/**
	 * Called from inside [serializeNBT] per slot, return true if you have written something into [nbt].
	 * If returning false, and slot does not contain an item, tag entry is discarded
	 */
	protected open fun attachSlotData(provider: Provider, slot: Int, nbt: CompoundTag, ops: RegistryOps<Tag>): Boolean {
		return false
	}

	protected open fun loadSlotData(provider: Provider, slot: Int, nbt: CompoundTag, ops: RegistryOps<Tag>) {

	}

	override fun serializeNBT(provider: Provider): CompoundTag {
		val ops = provider.createSerializationContext(NbtOps.INSTANCE)

		return CompoundTag().also {
			it["items"] = ListTag().also {
				for (i in 0 until size) {
					val tag = CompoundTag()
					var attached = attachSlotData(provider, i, tag, ops)

					if (items[i].isNotEmpty) {
						attached = true
						tag["item"] = ItemStack.OPTIONAL_CODEC.encodeStart(ops, items[i])
							.getOrThrow { RuntimeException("Unable to serialize item ${items[i]} at slot $i: $it") }
					}

					tag["slot"] = i

					if (attached) {
						it.add(tag)
					}
				}
			}
		}
	}

	override fun deserializeNBT(provider: Provider, nbt: CompoundTag) {
		val copy = observedItems.copyOf()
		items.fill(ItemStack.EMPTY)
		observedItems.fill(ItemStack.EMPTY)
		bitmap.clear()

		val seenSlots = IntOpenHashSet()
		val ops = provider.createSerializationContext(NbtOps.INSTANCE)

		if ("items" in nbt) {
			for (element in nbt.getList("items", Tag.TAG_COMPOUND.toInt())) {
				element as CompoundTag

				if ("slot" !in element) continue
				val slot = element.getInt("slot")
				if (!seenSlots.add(slot)) continue

				if ("item" in nbt) {
					ItemStack.OPTIONAL_CODEC.decode(ops, nbt["item"])
						.map { it.first }
						.ifError { LOGGER.error("Failed to deserialize item stack in slot $slot: ${it.message()}") }
						.ifSuccess {
							if (it.isNotEmpty) {
								items[slot] = it
								observedItems[slot] = it.copy()
								bitmap[slot] = true

								if (it.count != copy[slot].count || !ItemStack.isSameItemSameComponents(it, copy[slot]))
									notifySlotChanged(slot, copy[slot])
							}
						}
				}

				loadSlotData(provider, slot, element, ops)
			}
		}
	}

	open class Simple(size: Int) : EnhancedContainer<IContainerSlot>(size) {
		private val slots by lazy(LazyThreadSafetyMode.PUBLICATION) { Array(size) { IContainerSlot.Simple(it, this) } }

		override fun containerSlot(slot: Int): IContainerSlot {
			return slots[slot]
		}
	}

	open class WithListener(size: Int, private val listener: Runnable) : Simple(size) {
		override fun notifySlotChanged(slot: Int, old: ItemStack) {
			super.notifySlotChanged(slot, old)
			listener.run()
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
