package ru.dbotthepony.mc.otm.container

import it.unimi.dsi.fastutil.ints.IntCollection
import net.minecraft.world.Container
import net.minecraft.world.entity.player.Player
import net.minecraft.world.entity.player.StackedContents
import net.minecraft.world.inventory.StackedContentsCompatible
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Items
import net.minecraft.world.item.crafting.RecipeInput
import ru.dbotthepony.mc.otm.core.collect.any
import ru.dbotthepony.mc.otm.core.collect.filter
import ru.dbotthepony.mc.otm.core.collect.map
import ru.dbotthepony.mc.otm.core.isNotEmpty
import java.util.function.Predicate
import java.util.stream.Stream
import java.util.stream.StreamSupport

/**
 * "Backward-compatible" enhanced container interface, where all methods can be derived/emulated from existing [Container] code
 *
 * This is useful because it allows to interact with actually enhanced and regular containers through unified interface,
 * and actual implementations of this interface are likely to provide efficient method implementations in place of derived/emulated ones.
 */
interface IEnhancedContainer<out S : IContainerSlot> : Container, RecipeInput, Iterable<ItemStack>, StackedContentsCompatible {
	// https://youtrack.jetbrains.com/issue/KT-55080/Change-the-behavior-of-inheritance-delegation-to-delegates-implementing-Java-interfaces-with-default-methods
	override fun getMaxStackSize(): Int {
		return super.getMaxStackSize()
	}

	override fun startOpen(player: Player) {
		super.startOpen(player)
	}

	override fun stopOpen(player: Player) {
		super.stopOpen(player)
	}

	override fun canPlaceItem(slot: Int, itemStack: ItemStack): Boolean {
		return super.canPlaceItem(slot, itemStack)
	}

	override fun canTakeItem(container: Container, slot: Int, itemStack: ItemStack): Boolean {
		return super.canTakeItem(container, slot, itemStack)
	}

	override fun clearContent()
	override fun getContainerSize(): Int
	override fun getItem(slot: Int): ItemStack
	override fun removeItem(slot: Int, amount: Int): ItemStack
	override fun removeItemNoUpdate(slot: Int): ItemStack
	override fun setItem(slot: Int, itemStack: ItemStack)
	override fun setChanged()

	// provide non-ambiguous get and set operators
	operator fun get(slot: Int): ItemStack {
		return getItem(slot)
	}

	operator fun set(slot: Int, value: ItemStack) {
		setItem(slot, value)
	}

	fun containerSlot(slot: Int): S

	override fun fillStackedContents(contents: StackedContents) {
		forEach { contents.accountStack(it) }
	}

	/**
	 * Returns iterator over **all** slots this container has
	 */
	fun slotIterator(): Iterator<S> {
		return (0 until containerSize).iterator().map { containerSlot(it) }
	}

	fun nonEmptySlotIterator(): Iterator<S> {
		return nonEmptySlotIndexIterator().map { containerSlot(it) }
	}

	fun emptySlotIterator(): Iterator<S> {
		return emptySlotIndexIterator().map { containerSlot(it) }
	}

	fun nonEmptySlotIterator(allowedSlots: IntCollection): Iterator<S> {
		return nonEmptySlotIndexIterator(allowedSlots).map { containerSlot(it) }
	}

	fun emptySlotIterator(allowedSlots: IntCollection): Iterator<S> {
		return emptySlotIndexIterator(allowedSlots).map { containerSlot(it) }
	}

	private fun slotIndexWalker(allowedSlots: IntCollection, predicate: Predicate<ItemStack>): IntIterator {
		return object : IntIterator() {
			private val parent = allowedSlots.intIterator()
			private var foundNext = false
			private var next = -1

			private fun findNext() {
				if (!foundNext) {
					foundNext = true
					next = -1

					while (parent.hasNext()) {
						val i = parent.nextInt()

						if (predicate.test(this@IEnhancedContainer[i])) {
							next = i
							break
						}
					}
				}
			}

			override fun nextInt(): Int {
				findNext()

				if (next == -1)
					throw NoSuchElementException()

				foundNext = false
				val next = next
				this.next = -1
				return next
			}

			override fun hasNext(): Boolean {
				findNext()
				return next != -1
			}
		}
	}

	fun emptySlotIndexIterator(allowedSlots: IntCollection): IntIterator {
		return slotIndexWalker(allowedSlots) { it.isEmpty }
	}

	fun nonEmptySlotIndexIterator(allowedSlots: IntCollection): IntIterator {
		return slotIndexWalker(allowedSlots) { it.isNotEmpty }
	}

	fun emptySlotIndexIterator(): IntIterator {
		return emptySlotIndexIterator(slotRange)
	}

	fun nonEmptySlotIndexIterator(): IntIterator {
		return nonEmptySlotIndexIterator(slotRange)
	}

	fun slotIndexWithItemIterator(item: Item): IntIterator {
		return slotIndexWithItemIterator(item, slotRange)
	}

	fun slotIndexWithItemIterator(item: Item, allowedSlots: IntCollection): IntIterator {
		val parent = nonEmptySlotIndexIterator(allowedSlots).filter { this[it].isNotEmpty && this[it].item === item }

		return object : IntIterator() {
			override fun nextInt(): Int {
				return parent.next()
			}

			override fun hasNext(): Boolean {
				return parent.hasNext()
			}
		}
	}

	fun slotWithItemIterator(item: Item): Iterator<S> {
		return slotWithItemIterator(item, slotRange)
	}

	fun slotWithItemIterator(item: Item, allowedSlots: IntCollection): Iterator<S> {
		return slotIndexWithItemIterator(item, allowedSlots).map { containerSlot(it) }
	}

	fun nextEmptySlot(startIndex: Int): Int {
		for (i in startIndex until containerSize) {
			if (this[i].isEmpty) {
				return i
			}
		}

		return -1
	}

	fun nextNonEmptySlot(startIndex: Int): Int {
		for (i in startIndex until containerSize) {
			if (this[i].isNotEmpty) {
				return i
			}
		}

		return -1
	}

	fun nextSlotWithItem(startIndex: Int, item: Item): Int {
		var find = nextNonEmptySlot(startIndex)

		while (find != -1 && this[find].item !== item)
			find = nextNonEmptySlot(find + 1)

		return find
	}

	fun setChanged(slot: Int) {
		return setChanged()
	}

	fun getMaxStackSize(slot: Int, itemStack: ItemStack): Int {
		return maxStackSize
	}

	/**
	 * Returns iterator over **non-empty** [ItemStack]s inside this container
	 */
	override fun iterator(): Iterator<ItemStack> {
		return (0 until containerSize).iterator().map { this[it] }.filter { it.isNotEmpty }
	}

	override fun isEmpty(): Boolean {
		return nextNonEmptySlot(0) != -1
	}

	val hasEmptySlots: Boolean get() {
		return nextEmptySlot(0) != -1
	}

	override fun size(): Int {
		return containerSize
	}

	override fun countItem(item: Item): Int {
		var count = 0

		for (slot in slotWithItemIterator(item))
			count += slot.item.count

		return count
	}

	override fun hasAnyOf(items: Set<Item>): Boolean {
		if (Items.AIR in items && hasEmptySlots)
			return true

		return iterator().any { it.item in items }
	}

	override fun hasAnyMatching(predicate: Predicate<ItemStack>): Boolean {
		if (predicate.test(ItemStack.EMPTY) && hasEmptySlots)
			return true

		return iterator().any(predicate)
	}

	fun toList(): MutableList<ItemStack> {
		val list = ArrayList<ItemStack>(containerSize)

		for (i in 0 until containerSize) {
			list.add(this[i])
		}

		return list
	}

	fun copyToList(): MutableList<ItemStack> {
		val list = ArrayList<ItemStack>(containerSize)

		for (i in 0 until containerSize) {
			list.add(this[i].copy())
		}

		return list
	}

	private fun addItem(stack: ItemStack, simulate: Boolean, filterPass: Boolean, slots: IntCollection, onlyIntoExisting: Boolean, popTime: Int?, ignoreFilters: Boolean): ItemStack {
		if (stack.isEmpty || slots.isEmpty())
			return stack

		// двигаем в одинаковые слоты
		for (slot in slotWithItemIterator(stack.item, slots)) {
			val condition: Boolean

			if (slot is IFilteredContainerSlot) {
				condition = (ignoreFilters || !slot.isForbiddenForAutomation) &&
					ItemStack.isSameItemSameComponents(slot.item, stack) &&
					(ignoreFilters || !filterPass && !slot.hasFilter || filterPass && slot.hasFilter && slot.testSlotFilter(stack))
			} else {
				condition = (ignoreFilters || !filterPass) && ItemStack.isSameItemSameComponents(slot.item, stack)
			}

			if (condition) {
				val slotLimit = slot.maxStackSize(slot.item)

				if (slot.item.count < slotLimit) {
					val newCount = (slot.item.count + stack.count).coerceAtMost(slotLimit)
					val diff = newCount - slot.item.count

					if (!simulate) {
						slot.item.count = newCount
						slot.setChanged()

						if (popTime != null) {
							slot.item.popTime = popTime
						}
					}

					stack.shrink(diff)

					if (stack.isEmpty)
						return ItemStack.EMPTY
				}
			}
		}

		if (!onlyIntoExisting) {
			for (slot in emptySlotIterator(slots)) {
				val condition: Boolean

				if (slot is IFilteredContainerSlot) {
					condition = (ignoreFilters || !slot.isForbiddenForAutomation) &&
						(ignoreFilters || !filterPass && !slot.hasFilter || filterPass && slot.hasFilter && slot.testSlotFilter(stack))
				} else {
					condition = ignoreFilters || !filterPass
				}

				if (condition) {
					val diff = stack.count.coerceAtMost(slot.maxStackSize(stack))

					if (!simulate) {
						val copyToPut = stack.copy()
						copyToPut.count = diff
						slot.item = copyToPut

						if (popTime != null) {
							copyToPut.popTime = popTime
						}
					}

					stack.shrink(diff)

					if (stack.isEmpty)
						return ItemStack.EMPTY
				}
			}
		}

		return stack
	}

	/**
	 * Hint used internally by [IEnhancedContainer] to potentially speed up default method implementations
	 */
	val hasFilterableSlots: Boolean
		get() = false

	fun addItem(stack: ItemStack, simulate: Boolean, slots: IntCollection = slotRange, onlyIntoExisting: Boolean = false, popTime: Int? = null, ignoreFilters: Boolean = false): ItemStack {
		if (stack.isEmpty || slots.isEmpty())
			return stack

		if (ignoreFilters || !hasFilterableSlots) {
			return addItem(stack.copy(), simulate, filterPass = true, slots, onlyIntoExisting = onlyIntoExisting, popTime = popTime, ignoreFilters = true)
		} else {
			var copy = addItem(stack.copy(), simulate, filterPass = true, slots, onlyIntoExisting = onlyIntoExisting, popTime = popTime, ignoreFilters = false)
			copy = addItem(copy, simulate, filterPass = false, slots, onlyIntoExisting = onlyIntoExisting, popTime = popTime, ignoreFilters = false)
			return copy
		}
	}

	/**
	 * Unlike [addItem], modifies original [stack]
	 *
	 * @return Whenever [stack] was modified
	 */
	fun consumeItem(stack: ItemStack, simulate: Boolean, slots: IntCollection = slotRange, onlyIntoExisting: Boolean = false, popTime: Int? = null, ignoreFilters: Boolean = false): Boolean {
		if (stack.isEmpty)
			return false

		val result = addItem(stack, simulate = simulate, slots = slots, onlyIntoExisting = onlyIntoExisting, popTime = popTime, ignoreFilters = ignoreFilters)
		if (!simulate) stack.count = result.count
		return result.count != stack.count
	}

	fun fullyAddItem(stack: ItemStack, slots: IntCollection = slotRange, onlyIntoExisting: Boolean = false, popTime: Int? = null, ignoreFilters: Boolean = false): Boolean {
		if (!addItem(stack, simulate = true, slots = slots, popTime = popTime, onlyIntoExisting = onlyIntoExisting, ignoreFilters = ignoreFilters).isEmpty)
			return false

		return addItem(stack, simulate = false, slots = slots, popTime = popTime, onlyIntoExisting = onlyIntoExisting, ignoreFilters = ignoreFilters).isEmpty
	}

	fun stream(): Stream<ItemStack> {
		return StreamSupport.stream(spliterator(), false)
	}

	private class Wrapper(private val parent: Container) : IEnhancedContainer<IContainerSlot> {
		override fun containerSlot(slot: Int): IContainerSlot {
			return IContainerSlot.Simple(slot, parent)
		}

		override fun clearContent() {
			return parent.clearContent()
		}

		override fun setChanged() {
			return parent.setChanged()
		}

		override fun getContainerSize(): Int {
			return parent.containerSize
		}

		override fun getItem(slot: Int): ItemStack {
			return parent.getItem(slot)
		}

		override fun removeItem(slot: Int, amount: Int): ItemStack {
			return parent.removeItem(slot, amount)
		}

		override fun removeItemNoUpdate(slot: Int): ItemStack {
			return parent.removeItemNoUpdate(slot)
		}

		override fun setItem(slot: Int, itemStack: ItemStack) {
			return parent.setItem(slot, itemStack)
		}

		override fun stillValid(player: Player): Boolean {
			return parent.stillValid(player)
		}

		override fun getMaxStackSize(): Int {
			return parent.maxStackSize
		}

		override fun startOpen(player: Player) {
			parent.startOpen(player)
		}

		override fun stopOpen(player: Player) {
			parent.stopOpen(player)
		}

		override fun canPlaceItem(slot: Int, itemStack: ItemStack): Boolean {
			return parent.canPlaceItem(slot, itemStack)
		}

		override fun canTakeItem(container: Container, slot: Int, itemStack: ItemStack): Boolean {
			return parent.canTakeItem(container, slot, itemStack)
		}

		override fun isEmpty(): Boolean {
			return parent.isEmpty
		}

		override fun hasAnyMatching(predicate: Predicate<ItemStack>): Boolean {
			return parent.hasAnyMatching(predicate)
		}

		override fun hasAnyOf(items: Set<Item>): Boolean {
			return parent.hasAnyOf(items)
		}

		override fun countItem(item: Item): Int {
			return parent.countItem(item)
		}
	}

	companion object {
		fun wrap(other: Container): IEnhancedContainer<*> {
			if (other is IEnhancedContainer<*>)
				return other

			return Wrapper(other)
		}
	}
}
